import os
import re
import string
from pathlib import Path
from typing import List

import libcst as cst

from .cst import BaseCST
from .utils import validation

base_cst = BaseCST()


class ClassGenerator(BaseCST):
    def __init__(self, data: List[dict]) -> None:
        self.classname: str = data.get('name')
        self.description: str = data.get('description')
        self.base_classes: List[str] = data.get('base_classes')
        self.class_attributes: dict = data.get('attributes')
        self.instance_attributes: dict = data.get('instance_attributes')
        self.instance_methods: dict = data.get('instance_methods')

    @property
    def class_name(self) -> str:
        name = re.sub(r"[^a-zA-Z0-9 ]", "", self.classname)
        name = name.translate({ord(c): None for c in string.whitespace})
        return name.capitalize()

    @property
    def class_description_CST(self) -> List[cst.SimpleStatementLine]:
        description: str = f"'''\n    {self.description}\n    '''"
        body = cst.Expr(value=cst.SimpleString(description))
        return [cst.SimpleStatementLine(body=[body])]

    @property
    def bases_class_CST(self) -> List[cst.Arg]:
        bases: List[cst.Arg] = []
        for base_class in self.base_classes:
            arg = cst.Arg(cst.Name(str(base_class)))
            bases.append(arg)
        return bases

    @property
    def class_attributes_CST(self) -> List[cst.SimpleStatementLine]:
        fields: List[cst.SimpleStatementLine] = []
        for attr in self.class_attributes:
            ann_assign = self.create_class_attr_CST(
                str(attr.get('name')), str(attr.get('value')), str(attr.get('type')))
            fields.append(ann_assign)
        return fields

    @property
    def instance_attributes_CST(self) -> List[cst.FunctionDef]:
        class_method_attrs: List[cst.SimpleStatementLine] = []
        method_name: cst.Name = cst.Name(self.class_name.lower() + '_method')

        for attr in self.instance_attributes:
            class_method_attrs.append(
                self.create_instance_attr_CST(str(attr.get('name')), str(attr.get('value')))
            )
        body = cst.IndentedBlock(body=class_method_attrs)
        return [cst.FunctionDef(
            name=method_name,
            params=cst.Parameters(params=[cst.Param(cst.Name('self'))]),
            body=body,
            decorators=[],
            returns=None
        )]

    @property
    def class_getter_methods_CST(self) -> List[cst.FunctionDef]:
        getters: List[cst.FunctionDef] = []
        for attr in self.instance_attributes:
            getters.append(self.create_getter_CST(str(attr.get('name')), str(attr.get('type'))))
        return getters

    @property
    def class_setter_methods_CST(self) -> List[cst.FunctionDef]:
        setters: List[cst.FunctionDef] = []
        for attr in self.instance_attributes:
            setters.append(self.create_setter_CST(str(attr.get('name')), str(attr.get('type'))))
        return setters

    @property
    def instance_methods_CST(self) -> List[cst.FunctionDef]:
        methods_cst: List[cst.FunctionDef] = []
        for method in self.instance_methods:
            if method.get('definition'):
                method_cst = self.create_instance_method_CST(
                    method['definition'].get('name'),
                    method['definition'].get('decorators'),
                    method['definition'].get('arguments'),
                    method['definition'].get('statements'),
                    method['definition'].get('return_type')
                )
            else:
                method_cst = self.get_node_from_git(
                    method['github'].get('url'),
                    method['github'].get('filename'),
                    method['github'].get('target'),
                    True
                )
            if method_cst:
                methods_cst.append(method_cst)
        return methods_cst

    def make_cst(self) -> cst.ClassDef:
        class_name = cst.Name(self.class_name)
        class_body = (
            self.class_description_CST + self.class_attributes_CST
            + self.instance_attributes_CST + self.class_getter_methods_CST
            + self.class_setter_methods_CST + self.instance_methods_CST
        )
        base_classes = self.bases_class_CST
        return cst.ClassDef(
            name=class_name,
            body=cst.IndentedBlock(body=class_body),
            bases=base_classes
        )


class ImportGenerator(BaseCST):
    def make_cst(self, data: List[str]) -> List[cst.SimpleStatementLine]:
        imports_cst: List[cst.SimpleStatementLine] = []
        for statement in data:
            import_CST = self.create_import_CST(statement)
            imports_cst.append(import_CST)
        return imports_cst


class DataStructureGenerator(BaseCST):
    def make_cst(self, data: List[dict]) -> List[cst.Element]:
        data_structres_cst = []
        for d in data:
            data_structres_cst.append(self.create_data_structure_CST(d['name'], d['expression']))
        return data_structres_cst


class FunctionGenerator(BaseCST):

    def make_cst(self, data: List[dict]) -> cst.FunctionDef:
        functions_cst = []
        for f_data in data:
            if f_data.get('definition'):
                f_cst = self.create_function_CST(
                    f_data['definition'].get('name'),
                    f_data['definition'].get('decorators'),
                    f_data['definition'].get('arguments'),
                    f_data['definition'].get('statements'),
                    f_data['definition'].get('return_type')
                )
            else:
                f_cst = self.get_node_from_git(
                    f_data['github'].get('url'),
                    f_data['github'].get('filename'),
                    f_data['github'].get('target'),
                    False
                )
            if f_cst:
                functions_cst.append(f_cst)
        return functions_cst


def _make_classes_CST(data: List[dict]) -> List[cst.ClassDef]:
    classes_cst = []
    for cls_d in data:
        if cls_d:
            class_gen = ClassGenerator(cls_d)
            classes_cst.append(class_gen.make_cst())
    return classes_cst


def _make_module(module_body: list) -> cst.Module:
    return cst.Module(body=module_body, header=base_cst.default_header())


def _generate_code(module: cst.Module, dest_path: Path) -> None:
    code = cst.Module([]).code_for_node(module)
    with open(dest_path, 'w') as file:
        file.write(code)
    return code


def main(data: dict, dest_path: Path) -> str:
    """
    :dest_path - this is a path that the generated *.py file should be saved.
    :data - this is a dictionary that has the class components
    for example,
    {
        "ex1.py": {
            "classes": [{
                "name": "test",
                "description": "This is a test class.",
                "base_classes": ["ABC"],
                "attributes": [
                    {
                        "name": "a",
                        "type": "int",
                        "value": "3"
                    },
                    {
                        "name": "b",
                        "type": "str",
                        "value": "4"
                    },
                    {
                        "name": "c",
                        "type": "bool",
                        "value": "True"
                    }
                ],
                "instance_attributes": [{
                    "name": "test",
                    "value": "None",
                    "type": "str"
                }],
                "instance_methods": [{
                    "github": {
                        "url": "",
                        "filename": "",
                        "target": ""
                    },
                    "definition": {
                        "decorators": ["staticmethod"],
                        "name": "abc",
                        "arguments": [
                            {
                                "name": "a",
                                "type": "int",
                                "value": "1"
                            },
                            {
                                "name": "b",
                                "type": "str",
                                "value": "None"
                            }
                        ],
                        "statements": ["a = b", "return a"],
                        "return_type": "str"
                    }
                }]
            }],
            "functions": [{
                "github": {
                    "url": "",
                    "filename": "",
                    "target": ""
                },
                "definition": {
                    "decorators": ["my_decorator"],
                    "name": "abc",
                    "arguments": [
                        {
                            "name": "a",
                            "type": "int",
                            "value": "1"
                        },
                        {
                            "name": "b",
                            "type": "str",
                            "value": "None"
                        }
                    ],
                    "statements": ["a = b", "return a"],
                    "return_type": "str"
                }
            }],
            "imports": ["impor my_decorator", "from datetime import datetime"],
            "data_structures": [{"name": 'a', "expression": 99}, {"name": "b", "expression": [9, "hello", ["list2", 14, {}]]}]
        },
        "ex2.py": {
            "classes": [],
            "functions": [],
            "imports": [],
            "data_structures": []
        },
    }
    """
    is_valid, err_msg = validation(data)
    if not is_valid:
        raise Exception(err_msg)
    if not os.path.exists(dest_path):
        raise Exception('The dest_path you provide does not exist.')

    import_gen = ImportGenerator()
    data_structure_gen = DataStructureGenerator()
    func_gen = FunctionGenerator()

    for filename, filedata in data.items():
        dest_path = os.path.join(dest_path, filename)
        if not filename.lower().endswith('.py'):
            dest_path = dest_path.rglob("*.py")

        imports_cst: List[cst.SimpleStatementLine] = import_gen.make_cst(filedata['imports'])
        data_structures_cst: List[cst.SimpleStatementLine] = data_structure_gen.make_cst(filedata['data_structures'])
        functions_cst: List[cst.FunctionDef] = func_gen.make_cst(filedata['functions'])
        classes_cst: List[cst.ClassDef] = _make_classes_CST(filedata['classes'])

        module_body = imports_cst + data_structures_cst + classes_cst + functions_cst

        module = _make_module(module_body)
        code = _generate_code(module, dest_path)
        return code



import logging

FORMAT = '%(asctime)s %(clientip)-15s %(user)-8s %(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('pyclass_generator')


def validation(data: dict) -> tuple:
    err_msg = ''
    is_valid = True
    try:
        for filename, filedata in data.items():
            if not isinstance(filedata['classes'], list) or not isinstance(filedata['functions'], list) \
                or not isinstance(filedata['imports'], list) \
                    or not isinstance(filedata['data_structures'], list):
                is_valid = False
                err_msg = 'The classes or functions or imports or data_structures should be a list'
                break
            for class_data in filedata['classes']:
                if class_data['name']:
                    if not isinstance(class_data['name'], str):
                        is_valid = False
                        err_msg = 'class name should be a str.'
                        break
                    if not isinstance(class_data['description'], str):
                        is_valid = False
                        err_msg = 'class description should be a str.'
                        break
                    if not isinstance(class_data['base_classes'], list):
                        is_valid = False
                        err_msg = 'class base_classes should be a list.'
                        break
                    else:
                        for base_class in class_data['base_classes']:
                            if not isinstance(base_class, str):
                                is_valid = False
                                err_msg = 'class base_class should be a str.'
                                break
                    if not isinstance(class_data['attributes'], list):
                        is_valid = False
                        err_msg = 'class attributes should be a list.'
                        break
                    else:
                        for attr in class_data['attributes']:
                            if not isinstance(attr, dict):
                                is_valid = False
                                err_msg = 'class attribute should be a dict.'
                                break
                            else:
                                if sorted(attr.keys()) != sorted(['name', 'type', 'value']):
                                    is_valid = False
                                    err_msg = 'class attribute should have the name, type and value as the element key'
                                    break
                                else:
                                    for key, val in attr.items():
                                        if not isinstance(val, str):
                                            is_valid = False
                                            err_msg = 'class attribute should have a str as the element value'
                                            break

                    if not isinstance(class_data.get('instance_attributes'), list):
                        is_valid = False
                        err_msg = 'The class instance_attributes should be a list.'
                        break
                    else:
                        for attr in class_data['instance_attributes']:
                            if not isinstance(attr, dict):
                                is_valid = False
                                err_msg = 'class instance_attribute should be a dict.'
                                break
                            else:
                                if sorted(attr.keys()) != sorted(['name', 'type', 'value']):
                                    is_valid = False
                                    err_msg = 'class instance_attribute should have the name, type and value as the element key'
                                    break
                                else:
                                    for key, val in attr.items():
                                        if not isinstance(val, str):
                                            is_valid = False
                                            err_msg = 'class instance_attribute should have a str as the element value'
                                            break

                    if not isinstance(class_data['instance_methods'], list):
                        is_valid = False
                        err_msg = 'The class instance_methods should be a list.'
                        break
                    else:
                        for method in class_data['instance_methods']:
                            if sorted(method.keys()) == sorted(['github', 'definition']):
                                is_valid = False
                                err_msg = f'a Class method could not have both the github and definition as the key: {method}.'
                                break
                            else:
                                if method.get('github'):
                                    if not isinstance(method['github'], dict):
                                        is_valid = False
                                        err_msg = 'class github data should be a dict.'
                                        break
                                    else:
                                        if sorted(method['github'].keys()) == sorted(['url', 'filename', 'target']):
                                            for key, val in method['github'].items():
                                                if not isinstance(val, str):
                                                    is_valid = False
                                                    err_msg = 'class github every element should be a str.'
                                                    break
                                        else:
                                            is_valid = False
                                            err_msg = 'class github should have the url, filename and target as the key'
                                            break
                                elif method.get('definition'):
                                    if not isinstance(method['definition'], dict):
                                        is_valid = False
                                        err_msg = 'class instance method definition should be a dict.'
                                        break
                                    else:
                                        if sorted(method['definition'].keys()) == sorted(['decorators', 'name', 'arguments', 'statements', 'return_type']):
                                            if not isinstance(method['definition']['decorators'], list):
                                                is_valid = False
                                                err_msg = 'class instance method decorators should be a list'
                                                break
                                            else:
                                                for decorator in method['definition']['decorators']:
                                                    if not isinstance(decorator, str):
                                                        is_valid = False
                                                        err_msg = 'class instance method decorator should be a str'
                                                        break
                                            if not isinstance(method['definition']['name'], str):
                                                is_valid = False
                                                err_msg = 'class instance method name should be a str'
                                                break
                                            if not isinstance(method['definition']['arguments'], list):
                                                is_valid = False
                                                err_msg = 'class instance method arguments should be a list'
                                                break
                                            else:
                                                for arg in method['definition']['arguments']:
                                                    if sorted(arg.keys()) != sorted(['name', 'type', 'value']):
                                                        is_valid = False
                                                        err_msg = 'class instance method argument should have the name, type and value as the element key'
                                                        break
                                                    else:
                                                        for key, val in arg.items():
                                                            if not isinstance(val, str):
                                                                is_valid = False
                                                                err_msg = 'class instance method argument should have a str as the dict element.'
                                                                break
                                            if not isinstance(method['definition']['statements'], list):
                                                is_valid = False
                                                err_msg = 'class instance method statements should be a list'
                                                break
                                            else:
                                                for statement in method['definition']['statements']:
                                                    if not isinstance(statement, str):
                                                        is_valid = False
                                                        err_msg = 'class instance method statement should be a str'
                                                        break
                                            if not isinstance(method['definition']['return_type'], str):
                                                is_valid = False
                                                err_msg = 'class instance method return_type should be a str'
                                                break
                                        else:
                                            is_valid = False
                                            err_msg = 'class instance method definition should have the decorators, name, arguments, statements and return_type as the key'
                                            break
                                else:
                                    is_valid = False
                                    err_msg = f'A  Class instance method should have the github or definition as the key: {method}.'
                                    break
            for function_data in filedata['functions']:
                if sorted(function_data.keys()) == sorted(['github', 'definition']):
                    is_valid = False
                    err_msg = f'functions could not have both the github and definition as the key: {function_data}.'
                    break
                else:
                    if function_data.get('github'):
                        if not isinstance(function_data['github'], dict):
                            is_valid = False
                            err_msg = 'function github data should be a dict.'
                            break
                        else:
                            if sorted(function_data['github'].keys()) == sorted(['url', 'filename', 'target']):
                                for key, val in function_data['github'].items():
                                    if not isinstance(val, str):
                                        is_valid = False
                                        err_msg = 'function github every element should be a str.'
                                        break
                            else:
                                is_valid = False
                                err_msg = 'function github should have the url, filename and target as the key'
                                break
                    elif function_data.get('definition'):
                        if not isinstance(function_data['definition'], dict):
                            is_valid = False
                            err_msg = 'function definition should be a dict.'
                            break
                        else:
                            if sorted(function_data['definition'].keys()) == sorted(['decorators', 'name', 'arguments', 'statements', 'return_type']):
                                if not isinstance(function_data['definition']['decorators'], list):
                                    is_valid = False
                                    err_msg = 'function decorators should be a list'
                                    break
                                else:
                                    for decorator in function_data['definition']['decorators']:
                                        if not isinstance(decorator, str):
                                            is_valid = False
                                            err_msg = 'function decorator should be a str'
                                            break
                                if not isinstance(function_data['definition']['name'], str):
                                    is_valid = False
                                    err_msg = 'function name should be a str'
                                    break
                                if not isinstance(function_data['definition']['arguments'], list):
                                    is_valid = False
                                    err_msg = 'function arguments should be a list'
                                    break
                                else:
                                    for arg in function_data['definition']['arguments']:
                                        if sorted(arg.keys()) != sorted(['name', 'type', 'value']):
                                            is_valid = False
                                            err_msg = 'function argument should have the name, type and value as the element key'
                                            break
                                        else:
                                            for key, val in arg.items():
                                                if not isinstance(val, str):
                                                    is_valid = False
                                                    err_msg = 'function argument should have a str as the dict element.'
                                                    break
                                if not isinstance(function_data['definition']['statements'], list):
                                    is_valid = False
                                    err_msg = 'function statements should be a list'
                                    break
                                else:
                                    for statement in function_data['definition']['statements']:
                                        if not isinstance(statement, str):
                                            is_valid = False
                                            err_msg = 'function statement should be a str'
                                            break
                                if not isinstance(function_data['definition']['return_type'], str):
                                    is_valid = False
                                    err_msg = 'function return_type should be a str'
                                    break
                            else:
                                is_valid = False
                                err_msg = 'function definition should have the decorators, name, arguments, statements and return_type as the key'
                                break
                    else:
                        is_valid = False
                        err_msg = f'function should have the github or definition as the key: {function_data}.'
                        break
            for import_statement in filedata['imports']:
                if not isinstance(import_statement, str):
                    is_valid = False
                    err_msg = 'import statement should be a str'
                    break
            for data_structure in filedata['data_structures']:
                if sorted(data_structure.keys()) != sorted(['name', 'expression']):
                    is_valid = False
                    err_msg = 'data_structure should have the name and expression as the key'
                    break

    except Exception as e:
        is_valid = False
        err_msg = str(e)
    return is_valid, err_msg



from setuptools import setup

setup(
    name='pyclass_generator',
    version='0.1.4',
    description='This is a package to generate python class from a nested dictionary',
    url='https://github.com/Venus713/pyclass-generator',
    author='Venus713',
    author_email='userstar713@gmail.com',
    license='MIT',
    packages=[
        'pyclass_generator'
    ],
    install_requires=["libcst", "gitpython"],
    zip_safe=False
)



from abc import ABC
from pathlib import Path
from typing import Any, List, Optional  # Tuple, Dict

import git
import libcst as cst

from .utils import logger

# import libcst.matchers as m


class VisitorCollector(cst.CSTVisitor):
    def __init__(self, target_name: str = None, is_method: bool = False):
        self.target_name: str = target_name
        self.is_method: bool = is_method
        self.target_node: cst.FunctionDef = None

    def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
        return self.is_method

    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        if node.name.value == self.target_name:
            self.target_node = node
        return False


# class TransformerCollector(cst.CSTTransformer):
#     def __init__(self, *args, **kwargs):
#         pass

#     def leave_ClassDef(
#         self, original_node: cst.ClassDef, updated_node: cst.ClassDef
#     ) -> cst.CSTNode:
#         if m.matches(updated_node.name, m.Name()):
#             return updated_node.with_changes(
#                 name=cst.Name(
#                     self.class_name
#                 )
#             )
#         return updated_node

#     def leave_FunctionDef(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         log_stmt = cst.Expr(cst.parse_expression("print('returning')"))
#         return cst.FlattenSentinel([log_stmt, cst.Expr(cst.Newline()), updated_node])

#     def leave_Call(self, original_node: cst.Call, updated_node: cst.Call) -> cst.Call:
#         if m.matches(original_node.func, m.Name()):
#             return original_node.with_changes(
#                 func=cst.Name(
#                     "renamed_" + cst.ensure_type(original_node.func, cst.Name).value
#                 )
#             )
#         return original_node

#     def leave_Assign(self, old_node, updated_node):
#         log_stmt = cst.Expr(cst.parse_expression("print('returning')"))
#         return cst.FlattenSentinel([log_stmt, cst.Expr(cst.Newline()), updated_node])

#     def leave_AnnAssign(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         return updated_node

#     def leave_AsName(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         return updated_node

#     def leave_AugAssign(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         return updated_node

#     def leave_Decorator(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         return updated_node

#     def leave_SimpleStatementLine(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         if m.matches(original_node.body[0], m.Pass()):
#             attr_body = []
#             for f in self.fields:
#                 attr_body.append(cst.Expr(cst.parse_module(f)))
#             return original_node.with_changes(
#                 body=[
#                     attr_body
#                 ]
#             )
#         return original_node

#     def leave_SimpleStatementSuite(
#         self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
#     ) -> cst.CSTNode:
#         return updated_node


class GitRemoteProgress(git.remote.RemoteProgress):
    def update(self, op_code, cur_count, max_count=None, message=''):
        logger.info('update(%s, %s, %s, %s)' % (op_code, cur_count, max_count, message))


class BaseCST(ABC):

    def git_clone_to_temp(self, url: str) -> None:
        try:
            git.Repo.clone_from(url, './temp', progress=GitRemoteProgress())
        except git.GitCommandError as e:
            logger.info(str(e))
        except Exception as e:
            raise Exception(e)
        return

    def get_node_from_git(
        self, gitrepo_url: str, filename: str, target: str, is_method: bool = False
    ) -> cst.FunctionDef:
        self.git_clone_to_temp(gitrepo_url)
        target_file = Path(f'./temp/{filename}')
        if target_file.is_file():
            with open(f'./temp/{filename}') as f:
                source = f.read()
            source_tree = cst.parse_module(source)
            visitor = VisitorCollector(target, is_method)
            source_tree.visit(visitor)
            return visitor.target_node
        else:
            raise Exception(f'The file {filename} does not exist in the git repository.')

    def split_statement(
            self, statement: str, word: str, left: bool = False, right: bool = False) -> list:
        try:
            if left:
                return (statement.split(word, 2)[0]).strip()
            elif right:
                return (statement.split(word, 2)[1]).strip()
            else:
                return statement.split(word, 2)
        except Exception:
            raise Exception('Invailid Statement')

    def default_leading_lines(self) -> List[cst.EmptyLine]:
        return [cst.EmptyLine(
                indent=False, whitespace=cst.SimpleWhitespace(value='',),
                comment=None, newline=cst.Newline(value=None,),)]

    def default_comma(self) -> cst.Comma:
        return cst.Comma(
            whitespace_before=cst.SimpleWhitespace(value='',),
            whitespace_after=cst.SimpleWhitespace(value=' ',)
        )

    def default_equal(self) -> cst.AssignEqual:
        return cst.AssignEqual(
            whitespace_after=cst.SimpleWhitespace(value=' ',),
            whitespace_before=cst.SimpleWhitespace(value=' ',)
        )

    def default_header(self) -> List[cst.EmptyLine]:
        return [
            cst.EmptyLine(
                indent=True,
                whitespace=cst.SimpleWhitespace(value='',),
                comment=None,
                newline=cst.Newline(value=None,)
            )
        ]

    def create_target_CST(self, name: str) -> cst.AssignTarget:
        return cst.AssignTarget(target=cst.Name(value=name))

    def create_class_attr_CST(self, name: str, value: str, val_type: str = None) -> cst.SimpleStatementLine:
        attr_name = cst.Name(name)

        if val_type == 'dict':
            attr_value = cst.Dict(value)
        elif val_type == 'list':
            attr_value = cst.List(value)
        elif val_type == 'tuple':
            attr_value = cst.Tuple(value)
        elif val_type == 'float':
            attr_value = cst.Float(value)
        elif val_type == 'bool':
            attr_value = cst.Name(value)
        elif val_type == 'int' or (val_type == 'int' and value.isnumeric()):
            attr_value = cst.Integer(value)
        else:
            attr_value = cst.SimpleString(f"'{value}'")
        if val_type:
            annotation = cst.Annotation(
                annotation=cst.Name(val_type), whitespace_before_indicator=cst.SimpleWhitespace(value=''))
            return cst.SimpleStatementLine(
                body=[
                    cst.AnnAssign(
                        target=attr_name,
                        annotation=annotation,
                        value=attr_value,
                        equal=cst.AssignEqual(
                            whitespace_before=cst.SimpleWhitespace(value=' ',),
                            whitespace_after=cst.SimpleWhitespace(value=' ',)
                        )
                    )
                ],
                leading_lines=self.default_leading_lines()
            )
        else:
            return cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=(
                            cst.AssignTarget(
                                target=attr_name
                            )
                        ),
                        value=attr_value,
                        equal=cst.AssignEqual(
                            whitespace_before=cst.SimpleWhitespace(value=' ',),
                            whitespace_after=cst.SimpleWhitespace(value=' ',)
                        )
                    )
                ],
                leading_lines=self.default_leading_lines()
            )

    def create_instance_attr_CST(self, name: str, value: str) -> cst.SimpleStatementLine:
        targets = [
            cst.AssignTarget(
                target=cst.Attribute(
                    value=cst.Name('self'),
                    attr=cst.Name(name)
                )
            )
        ]
        value = cst.Name(value)
        return cst.SimpleStatementLine(body=[cst.Assign(targets=targets, value=value)])

    def create_getter_CST(self, name: str, return_type: str) -> cst.FunctionDef:
        getter_name: cst.Name = cst.Name(name)
        params: cst.Parameters = cst.Parameters(
            params=[
                cst.Param(cst.Name('self'), star=''),
                cst.Param(cst.Name(name), cst.Annotation(cst.Name(return_type)))
            ]
        )
        body: cst.IndentedBlock = cst.IndentedBlock(
            body=[
                cst.SimpleStatementLine(body=[
                    cst.Return(
                        value=cst.Attribute(
                            value=cst.Name('self'),
                            attr=cst.Name('_' + name)
                        )
                    )
                ])
            ]
        )
        decorators: List[cst.Decorator] = [
            cst.Decorator(
                decorator=cst.Name('property')
            )
        ]
        returns: cst.Annotation = cst.Annotation(
            annotation=cst.Name(return_type),
            whitespace_before_indicator=cst.SimpleWhitespace(value=' ',)
        )
        return cst.FunctionDef(
            name=getter_name, params=params, body=body, decorators=decorators,
            returns=returns, leading_lines=self.default_leading_lines()
        )

    def create_setter_CST(self, name: str, return_type: str) -> cst.FunctionDef:
        setter_name: cst.Name = cst.Name(name)
        params: cst.Parameters = cst.Parameters(
            params=[
                cst.Param(cst.Name('self'), star='', comma=self.default_comma()),
                cst.Param(cst.Name(name), cst.Annotation(
                    cst.Name(return_type),
                    whitespace_before_indicator=cst.SimpleWhitespace(value=' ',)), star=''
                )
            ]
        )
        body: cst.IndentedBlock = cst.IndentedBlock(
            body=[
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[
                                cst.AssignTarget(
                                    cst.Attribute(
                                        value=cst.Name('self'),
                                        attr=cst.Name('_' + name)
                                    )
                                )
                            ],
                            value=cst.Name(name)
                        )
                    ]
                )
            ]
        )
        decorators: List[cst.Decorator] = [
            cst.Decorator(cst.Attribute(value=cst.Name(name), attr=cst.Name('setter')))
        ]
        returns: cst.Annotation = cst.Annotation(
            annotation=cst.Name(return_type),
            whitespace_before_indicator=cst.SimpleWhitespace(value=' ',)
        )
        return cst.FunctionDef(
            name=setter_name, params=params, body=body, decorators=decorators,
            returns=returns, leading_lines=self.default_leading_lines()
        )

    def create_instance_method_CST(
        self,
        name: str,
        decorators: List[str],
        arguments: List[dict],
        statements: List[str],
        return_type: str
    ) -> cst.FunctionDef:
        method_name: cst.Name = cst.Name(name)
        if 'staticmethod' in decorators:
            params: List[cst.Param] = []
        elif classmethod in decorators:
            params: List[cst.Param] = [
                cst.Param(cst.Name('cls'), comma=self.default_comma(), star='')
            ]
        else:
            params: List[cst.Param] = [
                cst.Param(cst.Name('self'), comma=self.default_comma(), star='')
            ]
        for arg in arguments:
            if arg.get('type') == 'dict':
                default_value = cst.Dict(arg.get('value'))
            elif arg.get('type') == 'list':
                default_value = cst.List(arg.get('value'))
            elif arg.get('type') == 'tuple':
                default_value = cst.Tuple(arg.get('value'))
            elif arg.get('type') == 'float':
                default_value = cst.Float(str(arg.get('value')))
            elif arg.get('type') == 'int':
                default_value = cst.Integer(str(arg.get('value')))
            else:  # case of ['bool', 'str']
                default_value = cst.Name(str(arg.get('value')))
            params.append(
                cst.Param(
                    name=cst.Name(arg.get('name')),
                    annotation=cst.Annotation(
                        annotation=cst.Name(arg.get('type')),
                        whitespace_before_indicator=cst.SimpleWhitespace(value='',)
                    ),
                    default=default_value,
                    star='',
                    comma=self.default_comma(),
                    equal=self.default_equal()
                )
            )
        body: List[cst.SimpleStatementLine] = []
        for statement in statements:
            if '=' in statement:
                statement_name = self.split_statement(statement=statement, word='=', left=True)
                statement_value = self.split_statement(statement=statement, word='=', right=True)
                body.append(cst.SimpleStatementLine(body=[
                    cst.Assign(
                        targets=[cst.AssignTarget(cst.Name(statement_name))],
                        value=cst.Name(statement_value)
                    )
                ]))
            elif 'return' in statement:
                statement_value = self.split_statement(
                    statement=statement, word='return', right=True)
                body.append(cst.SimpleStatementLine(body=[
                    cst.Return(
                        value=cst.Name(statement_value),
                        whitespace_after_return=cst.SimpleWhitespace(value=' ')
                    )
                ]))
        decorators_CST: List[cst.Decorator] = []
        for decorator in decorators:
            decorators_CST.append(cst.Decorator(decorator=cst.Name(decorator)))
        returns: cst.Annotation = cst.Annotation(
            annotation=cst.Name(return_type),
            whitespace_before_indicator=cst.SimpleWhitespace(value=' ',),
            whitespace_after_indicator=cst.SimpleWhitespace(value=' ',)
        )
        return cst.FunctionDef(
            name=method_name,
            params=cst.Parameters(params=params),
            body=cst.IndentedBlock(body=body),
            decorators=decorators_CST,
            returns=returns,
            leading_lines=self.default_leading_lines()
        )

    def create_import_CST(self, statement: str = None) -> cst.SimpleStatementLine:
        return cst.parse_statement(statement)

    def create_data_structure_CST(self, name: str, value: Any) -> cst.SimpleStatementLine:
        expression = cst.parse_expression(str(value))
        target = self.create_target_CST(name)
        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[target],
                    value=expression
                )
            ]
        )

    def create_function_CST(
        self,
        name: str,
        decorators: List[str],
        arguments: List[dict],
        statements: List[str],
        return_type: str
    ) -> cst.FunctionDef:
        method_name: cst.Name = cst.Name(name)
        params: List[cst.Param] = []

        for arg in arguments:
            if arg.get('type') == 'dict':
                default_value = cst.Dict(arg.get('value'))
            elif arg.get('type') == 'list':
                default_value = cst.List(arg.get('value'))
            elif arg.get('type') == 'tuple':
                default_value = cst.Tuple(arg.get('value'))
            elif arg.get('type') == 'float':
                default_value = cst.Float(str(arg.get('value')))
            elif arg.get('type') == 'int':
                default_value = cst.Integer(str(arg.get('value')))
            else:  # case of ['bool', 'str']
                default_value = cst.Name(str(arg.get('value')))
            params.append(
                cst.Param(
                    name=cst.Name(arg.get('name')),
                    annotation=cst.Annotation(
                        annotation=cst.Name(arg.get('type')),
                        whitespace_before_indicator=cst.SimpleWhitespace(value='',)
                    ),
                    default=default_value,
                    star='',
                    comma=self.default_comma(),
                    equal=self.default_equal()
                )
            )
        body: List[cst.SimpleStatementLine] = []
        for statement in statements:
            body.append(cst.parse_statement(statement))
        decorators_CST: List[cst.Decorator] = []
        for decorator in decorators:
            decorators_CST.append(cst.Decorator(decorator=cst.Name(decorator)))
        returns: cst.Annotation = cst.Annotation(
            annotation=cst.Name(return_type),
            whitespace_before_indicator=cst.SimpleWhitespace(value=' ',),
            whitespace_after_indicator=cst.SimpleWhitespace(value=' ',)
        )
        return cst.FunctionDef(
            name=method_name,
            params=cst.Parameters(params=params),
            body=cst.IndentedBlock(body=body),
            decorators=decorators_CST,
            returns=returns,
            leading_lines=self.default_leading_lines()
        )
